/*
   File:             Commands_Functions.c
   Specifica:        funzioni di esecuzione comandi
*/

/* Librerie per chiamate sui file, directory e processi */
#include <sys/types.h>  
#include <sys/stat.h>   
#include <fcntl.h>

/* Libreria per le operazioni sulle directory */  
#include <dirent.h>
#include <time.h>

/* Librerie per i processi */  
#include <sys/wait.h>   

#include "Macros.h" 
#include "Utility_Function.h" 
#include "Commands_Function.h"





/* Implementazione della funzione che esegue comando esterno
   execvp: tramite la famiglia delle chiamate di sistema exec, un processo puo' cambiare il codice che sta eseguendo.
   Piu' precisamente, puo' rimpiazzare il suo codice, dati, heap e stack con quelli di un eseguibile. Da notare le finali:
   p : indica una chiamata che prende un nome di file come argomento e lo cerca nelle directory specficate nella variabile d'ambiente path
   l : indica una chiamata che riceve una lista di argomenti terminata da un puntatore NULL
   v : indica una chiamata che riceve un vettore nello stesso formato di argv[]
   e : indice una chiamata che riceve un vettore di ambiente envp[] invece di usare l'ambiente corrente
*/

int externCommand(unsigned int argc, char **argv)
{
    int stato;
    const int pid=fork();
    IFERROR_THREE(pid,"fork() raised and error! Impossible to create another process",exit(EXIT_FAILURE)); 
    if(pid)
    { 
        /* Siamo nel processo padre */
        wait(&stato);
        if( WIFEXITED(stato)) stato=WEXITSTATUS(stato);
        else stato=WTERMSIG(stato);
        return stato;
    } 
    else
    {
        /* Siamo nel processo figlio */
        execvp(argv[0],argv);
        printf("%s : command not found",argv[0]);
        exit(EXIT_FAILURE);
    }
}





/* Implementazione della funzione che esegue il comando echo */

int echo(unsigned int argc, char **argv)
{
    for(unsigned int i=1; i<argc; i++) printf("%s ",argv[i]);
    printf("\n");
    return 0;
}






/* Implementazione della funzione che esegue il comando ls
   Readdir legge tutti i file contenuti in una directory uno alla volta. Ad ogni lettura
   viene restituito un puntatore ad una struttura contenente le informazioni sul file 
   corrente. Alla lettura successiva all'ultimo file viene restituito NULL. Richidede 
   un puntatore a directory e torna un puntatore struct dirent dove sono contenute le informazioni
   sui file contenuti nella directory.
 */

int ls(unsigned int argc, char *argv[])
{
    DIR *dir;
    struct dirent *file;
    if(argc==1) argv[1]=".";
    else if (argc==2)
    {
        if (strcmp(argv[1],"-r")==0) return walk_recur(current_directory());
        else invalid_argument(argv[1]);
        return 0;
    }
    IFOPEN_DIR(dir=opendir(argv[1]),"Failed to open the directory!",return EXIT_FAILURE);
    while((file=readdir(dir))!=NULL)
    {
        if (!strcmp(file->d_name,".") || !strcmp(file->d_name,"..")) continue;
        puts(file->d_name);
    }
    IFERROR_TWO(closedir(dir),argv[1]);
    return 0;
}




/* Implementazione della funzione che esegue il comando help */

inline int help(unsigned int argc, char **argv)
{
    puts("Giuseppe Persico\nShell Unix");
    return 0;
}




/* Implementazione della funzione che esegue il comando pwd */

inline int pwd(unsigned int argc, char **argv)
{
    puts(current_directory());
    return 0;
}





/* Implementazione della funzione che esegue il comando cd */

int cd(unsigned int argc, char **argv)
{
    /* chdir è una funzione di sistema per cambiare directory corrente */
    /* getenv è una funzione di sistema per leggere il valore di variabili d'ambiente */
    if(argc==1)
    {
        IFERROR_THREE(chdir(getenv("HOME")),"Impossible to read the environment variable setted!", return 0);
    }
    else IFERROR_THREE(chdir(argv[1]), argv[1], return 0);
    return 0 ;
}





/* Implementazione della funzione che esegue il comando cat */

int cat(unsigned int argc, char **argv)
{
    if (argc==2)
    {
        FILE *fp;
        if ((fp=fopen(argv[1],"r"))==NULL) printf("%s : No such file or directory\n",argv[1]);
        else
        {
            char buffer[MAX_BUFFER_SIZE];
            fread(buffer,sizeof(buffer[0]),sizeof(buffer)/sizeof(buffer[0]),fp);
            for (unsigned int i=0; i<MAX_BUFFER_SIZE; i++) printf("%c",buffer[i]);
        }
    }
    else puts("usage: cat filename ...\n");
    return 0;
}





/* Implementazione della funzione che esegue il comando rm */

int rm(unsigned int argc, char **argv)
{
    if (argc>=2) for (unsigned int nfiles=1; nfiles<argc; nfiles++) check_and_remove(argv[nfiles]);
    else puts("Usage:\n rm file1 ... filen");
    return 0;
}





/* Implementazione della funzione che esegue il comando mv */

int mv(unsigned int argc, char **argv)
{
    if (argc<3) puts("Usage:\nmv file1 ... filen destination");
    else if (argc>=3 && !isDirectory(argv[argc-1])) puts("Destination must be a directory!");
    else if (argc>=3 && isDirectory(argv[argc-1]))
    {
        for (unsigned int nfiles=1; nfiles<argc-1; nfiles++)
        {
            char *newpath;
            if (isDirectory(argv[nfiles]))
            {
                newpath=linkThree(argv[argc-1],"/",argv[nfiles]);
                IFERROR_THREE(mkdir(newpath,S_IRWXU|S_IRWXG|S_IRWXO),"Error while creating the new directory!",return -1);
                analyze_cases(move_recur(argv[nfiles],newpath));
            }
            else 
            {
                copyfile(argv[nfiles],newpath=linkThree(argv[argc-1],"/",argv[nfiles]));
                IFERROR_THREE(remove(argv[nfiles]),"Error while removing a file!",return 0);
            }
        }
    }
    return 0;
}




/* Implementazione della funzione che esegue il comando cp */

int cp(unsigned int argc, char *argv[])
{
    if(argc<3) puts("Usage:\ncp file1 ... filen destination");
    else if (argc>=3 && !isDirectory(argv[argc-1])) puts("Destination must be a directory");
    else if (argc>=3 && isDirectory(argv[argc-1]))
    {
        for (unsigned int nfiles=1; nfiles<argc-1; nfiles++)
        {
            char *newpath;
            if (isDirectory(argv[nfiles]))
            {
                newpath=linkThree(argv[argc-1],"/",argv[nfiles]);
                IFERROR_THREE(mkdir(newpath,S_IRWXU|S_IRWXG|S_IRWXO),"Error while creating the new directory!",return -1);
                analyze_cases(copy_recur(argv[nfiles],newpath));
            }
            else copyfile(argv[nfiles],newpath=linkThree(argv[argc-1],"/",argv[nfiles]));
        }
    }
    return 0;
}




/* Implementazione della funzione che esegue il comando find */

int find(unsigned int argc, char **argv)
{
    if (argc<2) puts("Usage:\n find file_name search_path");
    else if (argc==2) analyze_cases(find_recur(argv[1],current_directory()));
    else if (argc==3 && isDirectory(argv[2])) analyze_cases(find_recur(argv[1],argv[2]));
    return 0;
}




/* Implementazione della funzione che esegue il comando diff */

int diff(unsigned int argc, char **argv)
{
    if (argc>3 || argc<2) puts("Usage:\n diff file1 file2");
    else
    {
        info_file(argv[1]);
        info_file(argv[2]);
    }
    return 0;
}




/* Implementazione delle funzione che esegue il comando grep */

int grep(unsigned int argc, char **argv)
{
    if (argc<3) puts("Usage:\n grep pattern file_name");
    else if (argc==3 && !isDirectory(argv[2]))
    {
        FILE *fp;
        if ((fp=fopen(argv[2],"r"))==NULL) printf("%s : No such file or directory\n",argv[1]);
        else
        {
            char buffer[MAX_BUFFER_SIZE];
            char *ptr;
            for (unsigned int i=0; i<MAX_BUFFER_SIZE; i++) if (fgets(buffer,MAX_BUFFER_SIZE,fp)!=NULL) if ((ptr=strstr(buffer,argv[1]))!=NULL) puts(buffer);
        }
    }
    return 0;
}