/*
 File:      Interpretation_Commands.c
 Specifica: funzioni di interpretazione dei comandi
 */

/* Include le librerie per chiamate sui file */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* Include le librerie per chiamate sui processi */
#include <sys/types.h>  
#include <sys/wait.h>   

#include <stdio.h>

#include "Utility_Function.h" 
#include "Macros.h" 
#include "Commands_Function.h"
#include "Interpretation_Commands.h"




/* Funzione di interpretazione di una sequenza di comandi.
   Per ogni comando nella sequenza, separo gli argomenti e interpreto il comando. Infine libero
   la memoria occupata.
*/

int interpretCommandSequence(unsigned int argc, char **argv)
{
    int stato;
    unsigned int newargc;
    char **newargv;
    for(unsigned int i=0; i<argc; i++ )
    {
        newargv = splitArgument(argv[i]," ",&newargc);
        stato = interpretCommandBackground(newargc, newargv);
        freeArgument(newargv);
    }
    return stato;
}





/* Funzione di interpretazione di una comando in background.
   Se non c'è la & che indica il background, richiamo la funzione di intepretazione che redirige l'output.
   Se il pid generato dalla fork e' zero, siamo nel figlio, dunque elimino la & commerciale ed eseguo la
   funzione normale di intepretazione. 
*/

int interpretCommandBackground(unsigned int argc, char **argv)
{
    int pid;
    if( (argc < 2) || strcmp(argv[argc-1],"&") != 0 ) return interpretCommandRedirection(argc, argv);
    IFERROR_TWO(pid=fork(),"fork");
    if(pid==0)
    { 
        argv[argc-1]=NULL;
        exit(interpretCommandRedirection(argc-1, argv));
    }
    else return 0;
}






/* Funzione di interpretazione di un comando con possibilita' di redirezione dello standard output
 
   dup() e dup2() creano una copia del descrittore di file passato come argomento. Se le chiamate tornano
   con successo, il vecchio e il nuovo descrittore possono essere usati intercambiabilmente. Essi condividono
   i lock (accesso in mutua esclusione), i puntatori alla posizione nel file, e i vari flag. 
 
   dup()  :  cerca la prima locazione vuota libera con indice piu' basso nella tabella dei descrittori di file e vi copia
   il contenuto della locazione occupata dal descrittore passato come argomento.
   dup2() :  chiude il nuovo descrittore passato come argomento se e' attivo e vi copia il contenuto della locazione occupata
   dal vecchi descrittore (sempre passato come argomento). Se entrambe hanno successo, restituiscono l'indice della
   locazione del nuovo descrittore di file, altrimenti restituiscono -1.
   FUNZIONAMENTO:
 
   1) Il processo shell padre si duplica tramite la fork (vedi sopra) e attende tramite wait che il figlio termini.
   2) Il processo shell figlio apre in lettura il file ls.out, creandolo o troncandolo a seconda dei casi.
   3) Il processo shell figlio duplica il descrittore di file ls.out (con dup2()) nel descrittore di standard output
      che ha assegnato numero 1) e poi chiude il descrittore di file originario di ls.out. Così tutto lo standard output viene
      rediretto verso ls.out.
   4) Il processo shell figlio, tramite una exec() esegue il codice associato al comando lanciato e che ridirigera' l'output.
      Visto che i descrittori di file sono conservati dalla exec(), tutto lo stdout viene rediretto su ls.out.
   5) Quando il processo shell figlio termina, il padre si risveglia. Il bello e' che i descrittori di file del padre non vengono
      toccati dal figlio in quanto ogni processo mantiene la sua tabella privata dei descrittori di file. 
*/

int interpretCommandRedirection(unsigned int argc, char **argv)
{
    /* Se non viene trovato il comando di redirezione, richiama interpreta() */
    if ((argc<3) || (strcmp(argv[argc-2],">")!=0) )  return interpretCommand(argc, argv);
    /* Apro file su cui redirigere */
    int fd;
    IFERROR_THREE(fd=open(argv[argc-1],O_WRONLY|O_CREAT|O_TRUNC,0644), argv[argc-1], return 0);
    /* Duplico lo standard output su un descrittore (stream) temporaneo */ 
    int temp=dup(STDOUT);
    /* Rimpiazzo lo standard output con fd che e' il descrittore del file aperto in precedenza */
    dup2(fd,STDOUT);
    /* Chiudo fd perche' non serve piu' */
    close(fd);
    /* Elimina i due token finali ">" e nome_file_su_cui_redirigere */
    argv[argc-2]=NULL;
    /* Interpretazione del comando */
    int res=interpretCommand(argc-2, argv);
    /* Ripristino lo standard output originario dal descrittore temporaneo */
    dup2(temp, STDOUT); 
    close(temp); 
    return res;
}








/* Funzione di interpretazione di un comando
 
   Gli indirizzi delle funzioni che implementano i comandi sono memorizzati in un vettore di puntatori
   a funzioni che ha come argomenti: un intero, un vettore di stringhe e torna un intero.
   E' possibile fare cio' perche' tutte le funzioni hanno gli stessi tipi di parametri e di tipo di ritorno.
   Da notare che l'ultimo elemento e' la funzione che implementa e richiama comandi esterni.
   Il comando immesso dunque: viene splittato in token (splitArgument). Viene chiamata la funzione "interpreta"
   che cerca il primo token (che sarebbe il nome del comando) nel vettore comandi. Dunque viene eseguita la
   funzione corrispondente al nome del comando. Se non trovo il comando tra i comandi che conosco, alla fine i 
   fara' riferimento  all'ultima delle funzioni, ovvero a quella che esegue un comando esterno. 
   Alla fine, eseguo la funzione i-esima con numero di argomenti argc e vettore degli  argomenti argv.
*/

int interpretCommand(unsigned int argc, char **argv)
{
    /* Vettore con i nomi dei comandi riconosciuti e con ultimo elemento NULL */
    char *commands_names[]={"cd","pwd","help","ls","cp","echo","cat","rm","mv","find","diff","grep",NULL};
    int (*functions_names[])(unsigned int, char **) = {cd,pwd,help,ls,cp,echo,cat,rm,mv,find,diff,grep,externCommand};
    if(argc==0) return 0;
    int i=0;
    while (commands_names[i]!=NULL)
    {
        if(strcmp(commands_names[i],argv[0])==0) break;
        else i++;
    }
    return functions_names[i](argc,argv); 
}









/* verifica se sono terminati processi che girano in background
   WIFEXITED(s)   : viene valutata ad un valore diverso da zero, se il figlio termina normalmente
   WEXITSTATUS(s) : restituisce il codice di ritorno codificato negli 8 bit meno significativi di s
   WNOHANG        : torna subito se nessun processo figlio e' terminato.
   WUNTRACED      : torna subito anche nel caso in cui i figli sono in stato di stop e il cui stato non e' stato riportato
   wait(s)        : sospende un processo fino alla terminazione di uno dei suoi figli. Se non ha figli torna subito -1
 
   waitpid(pid,s,options) : simile, solo che permette di specializzare l'attesa. Ovvero quale processo aspettare. pid puo' essere:
   < -1 aspetta la terminazione di un qualsiasi processo figlio il cui process ID group e' uguale al modulo di pid
   -1 aspetta la terminazione di un qualsiasi processo figlio. In questo caso e' uguale e wait()
   0 aspetta la terminazione di un processo figlio che ha il process ID group uguale a quello del chiamante
   > 0 aspetta la terminazione del processo figlio il cui PID e' uguale al valore passato
*/

void checkEndProcess()
{  
    int stato, pid;
    if( (pid=waitpid(-1,&stato, WNOHANG))>0)
    {
        if(WIFEXITED(stato)) stato=WEXITSTATUS(stato);
        else stato=WTERMSIG(stato);
    }
}