/*
   File:       Utility_Functions.c
   Specifica:  funzioni di utilita'
*/

/* Librerie per stat */
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "Utility_Function.h"
#include "Macros.h"




/* Ritorna 1 se il parametro passato corrisponde ad una directory ed e' possibile ottenere informazioni su di essa */

const inline int isDirectory(const char *filename)
{
    struct stat info;
    return (stat(filename, &info)!=-1) && S_ISDIR(info.st_mode);
}



/* Funzione che controlla la corretta rimozione di un file */

const inline int remove_error(const int to_remove)
{
    if (to_remove==-1) puts("Error removing a file! Permission denied!");
    return 0;
}



/* Ritorna una stringa ottenuta concatenando tre stringhe passate come parametri. */

char *linkThree(const char *str1, const char *str2, const char *str3)
{
    char *conc;
    check_alloc(conc=(char *)malloc(strlen(str1)+strlen(str2)+strlen(str3)+1));
    strcpy(conc,str1);
    strcat(conc,str2);
    return strcat(conc,str3);
}




/* Ritorna una stringa ottenuta concatenando le due stringhe passate come parametri. */

char *linkTwo(const char *str1, const char *str2)
{
    char *conc;
    check_alloc(conc=(char *)malloc(strlen(str1)+strlen(str2)+1));
    strcpy(conc,str1);
    return strcat(conc,str2);
}




/* splitArgument() ritorna  il  vettore dei token nella 
   stringa line separati da un qualsiasi carattere
   separator presente nella stringa sep, e scrive 
   in argc il numero di token trovati.
   strsep() : Estrae i token (da una stringa passata come parametro) divisi da qualsiasi
   carattere passato a "separatori", tornando un vettore di stringhe. Infine il numero
   di token trovati verra' salvato in "token_number".
   FUNZIONAMENTO:
   
   VARIABILI: i e' il numero dei token trovati, **argv e' il vettore dei token e linecopy e' una copia di linea
   Avviene l'allocazione e la copia di linea in copyline. Alloco il vettore dei token e strsep() estrae i token 
   dalla stringa passata come parametro salvandoli in argv. Chiaramente in questo ciclo, il vettore dei token dovra'
   essere riallocato ad ogni iterazione.
*/

char **splitArgument(char *line, char *separator, unsigned int *token_number)
{
    char **argv;       
    char *linecopy; 
    check_alloc(linecopy=(char *)malloc(strlen(line)+1));
    strcpy(linecopy,line);
    check_alloc(argv=(char **)malloc(sizeof(char *)));
    unsigned int i=0;
    while((argv[i]=strsep(&linecopy,separator))!=NULL) check_alloc(argv=(char **)realloc(argv,((++i)+1)*sizeof(char *)));
    *token_number=i;
    return argv;
}






/* Dealloca un vettore di stringhe denotato dall'argomento che deve essere un puntatore
   ritornato da "splitArgument". Ogni "splitArgument" deve avere una corrispondente 
   "freeArgument" per evitare sprechi di memoria.
   L'area a cui puntano gli argomenti di argv e' stata allocata con una sola malloc.
   Notare infatti linecopy=(char*)malloc(...)) in splitArgument. Infine viene deallocato
   completamente il vettore di puntatori.
*/

const inline void freeArgument(char **argv)
{
    free(argv[0]);
    free(argv);
}




/* Funzione di controllo della scrittura su stream */

inline void writeAndCheck(int destination, char **buf, const unsigned long len)
{
    int n;
    if ((n=(write(destination,*buf,len)==-1)))
    {
        puts("Failed to write commandBuffer on STDOUT");
        exit(EXIT_FAILURE);
    }
}



/* Funzione di controllo della lettura da buffer */

inline void readAndCheck(int source, char *buf, const unsigned long len)
{
    int n;
    if ((n=((int)read(STDIN,buf,MAX_COMMAND_SIZE)))==-1)
    {
        puts("Failed to read from commandBuffer");
        exit(EXIT_FAILURE);
    }
    buf[n-1]=0;
}





/* Funzione che copia due file. Viene richiamata dal comando copy */

const void copyfile(const char *namesrc, const char *namedest)
{
    int file_source, file_destination;
    long letti;
    char buffer[MAX_BUFFER_SIZE];
    IFERROR_THREE(file_source=open(namesrc,O_RDONLY), namesrc, return); 
    /* Codifica dei permessi in lettura e scrittura */
    IFERROR_THREE(file_destination=open(namedest,O_WRONLY|O_CREAT|O_TRUNC,0644), namedest, return);
    while((letti=read(file_source,buffer,MAX_BUFFER_SIZE))>0)
    {
        IFERROR_TWO(write(file_destination,buffer,letti), namedest);
    }
    IFERROR_TWO(letti,namesrc);
    IFERROR_TWO(close(file_destination), namedest);
}






/* Funzione che attraversa ricorsivamente una directory e tutte le sue sottodirectory
   La funzione lstat() restituisce informazioni sul file corrente e prende come parametri:
   const char *path : il percorso del file che si sta analizzando, puo' essere relativo o assoluto
   struct stat *buf	: un puntatore ad una struttura dove le informazioni verranno memorizzate.
   Torna -1 se fallisce. 
   st_mode invece e' un intero che codifica il tipo di file ed i diritti di accesso. E' possibile testare 
   tali informazioni con delle apposite macro, in questo caso, S_ISDIR dira' se è una directory o meno. 
   Sara' falso per i collegamenti a directory, in quanto link simbolici.
   In questa funzione verranno saltati durante la stampa a video, la directory corrente, madre e i file nascosti
*/

const int walk_recur(const char *namesrc)
{
	struct dirent *file;
	DIR *dir;
	struct stat st;
	char buffer[FILENAME_MAX];
	int len=(int)strlen(namesrc);
    
	if (len>=FILENAME_MAX-1) return WALK_NAMETOOLONG;
	strcpy(buffer,namesrc);
	buffer[len++]='/';
    if ((dir=opendir(namesrc))==NULL) return WALK_OPEN_ERROR;
    while ((file=readdir(dir)))
    {
		if (!strcmp(file->d_name,".") || !strcmp(file->d_name,"..") || file->d_name[0]=='.') continue;
		strncpy(buffer+len,file->d_name,FILENAME_MAX-len);
        IFERROR_THREE(lstat(buffer,&st),"Can't obtain information about last file or directory!",continue);
		if (S_ISDIR(st.st_mode)) walk_recur(buffer);
		puts(buffer);
	}
	closedir(dir);
	return WALK_OK;
}




/* Funzione che attraversa ricorsivamente una directory e le sue sottodirectory copiando tutto il contenuto */

const int copy_recur(const char *namesrc, const char *namedest)
{
    struct dirent *file;
	DIR *dir;
	struct stat st;
	char buffer[FILENAME_MAX];
	int len=(int)strlen(namesrc);
    char *newpath;
    
	if (len>=FILENAME_MAX-1) return WALK_NAMETOOLONG;
	strcpy(buffer,namesrc);
	buffer[len++]='/';
    if ((dir=opendir(namesrc))==NULL) return WALK_OPEN_ERROR;
	while ((file=readdir(dir)))
    {
		if (!strcmp(file->d_name,".") || !strcmp(file->d_name,"..") || file->d_name[0]=='.') continue;
		strncpy(buffer+len, file->d_name,FILENAME_MAX-len);
        IFERROR_THREE(lstat(buffer,&st),"Can't obtain information about last file or directory!",continue);
		if (S_ISDIR(st.st_mode))
        {
            newpath=linkThree(namedest,"/",file->d_name);
            if (mkdir(newpath,S_IRWXU|S_IRWXG|S_IRWXO)==-1) return WALK_MKDIR_ERROR;
            copy_recur(buffer,newpath);
        }
        else copyfile(buffer,newpath=linkThree(namedest,"/",file->d_name));
	}
	closedir(dir);
    return WALK_OK;
}





/* Funzione che attraversa ricorsivamente una directory e le sue sottodirectory spostando tutto in un'altra
   directory ed eliminando tutto dalla directory sorgente.
*/

const int move_recur(const char *namesrc, const char *namedest)
{
    struct dirent *file;
	DIR *dir;
	struct stat st;
	char buffer[FILENAME_MAX];
	int len=(int)strlen(namesrc);
    char *newpath;
    
	if (len>=FILENAME_MAX-1) return WALK_NAMETOOLONG;
	strcpy(buffer,namesrc);
	buffer[len++]='/';
    if ((dir=opendir(namesrc))==NULL) return WALK_OPEN_ERROR;
	while ((file=readdir(dir)))
    {
		if (!strcmp(file->d_name,".") || !strcmp(file->d_name,"..")) continue;
		strncpy(buffer+len, file->d_name,FILENAME_MAX-len);
        IFERROR_THREE(lstat(buffer,&st),"Can't obtain information about last file or directory!",continue);
		if (S_ISDIR(st.st_mode))
        {
            newpath=linkThree(namedest,"/",file->d_name);
            if (mkdir(newpath,S_IRWXU|S_IRWXG|S_IRWXO)==-1) return WALK_MKDIR_ERROR;
            move_recur(buffer,newpath);
        }
        else
        {
            copyfile(buffer,newpath=linkThree(namedest,"/",file->d_name));
            if (remove(buffer)==-1) return WALK_REMOVE_FILE;
        }
        rmdir(buffer);
	}
    closedir(dir);
    rmdir(namesrc);
	return WALK_OK;
}




/* Funzione che attravera ricorsivamente una directory e le sue sottodirectory eliminando tutto */

const int remove_recur(const char *namesrc)
{
    struct dirent *file;
	DIR *dir;
	struct stat st;
	char buffer[FILENAME_MAX];
	int len=(int)strlen(namesrc);
    
	if (len>=FILENAME_MAX-1) return WALK_NAMETOOLONG;
	strcpy(buffer,namesrc);
	buffer[len++]='/';
    if ((dir=opendir(namesrc))==NULL) return WALK_OPEN_ERROR;
	while ((file=readdir(dir)))
    {
		if (!strcmp(file->d_name, ".") || !strcmp(file->d_name, "..")) continue;
		strncpy(buffer+len, file->d_name,FILENAME_MAX-len);
        IFERROR_THREE(lstat(buffer,&st),"Can't obtain information about last file or directory!",continue);
		if (S_ISDIR(st.st_mode)) remove_recur(buffer);
        else
        {
            if (remove(buffer)==-1) return WALK_REMOVE_FILE;
        }
        rmdir(buffer);
	}
	closedir(dir);
    rmdir(namesrc);
    return WALK_OK;
}




/* Funzione che attraversa ricorsivamente una directory e sottodirectory per trovare un pattern in input */

const int find_recur(const char *namesrc, const char *namedest)
{
    struct dirent *file;
	DIR *dir;
    char *ptr;
	struct stat st;
	char buffer[FILENAME_MAX];
	int len=(int)strlen(namesrc);
    
	if (len>=FILENAME_MAX-1) return WALK_NAMETOOLONG;
	strcpy(buffer,namesrc);
	buffer[len++]='/';
    if ((dir=opendir(namesrc))==NULL) return WALK_OPEN_ERROR;
	while ((file=readdir(dir)))
    {
		if (!strcmp(file->d_name, ".") || !strcmp(file->d_name, "..") || file->d_name[0]=='.') continue;
		strncpy(buffer+len, file->d_name,FILENAME_MAX-len);
        IFERROR_THREE(lstat(buffer,&st),"Can't obtain information about last file or directory!",continue);
		if (S_ISDIR(st.st_mode)) find_recur(buffer,namedest);
        if ((ptr=strstr(namesrc,namedest))!=NULL) puts(buffer);
	}
	closedir(dir);
	return WALK_OK;
}




/* Funzione che analizza i vari esiti della discesa ricorsiva all'interno delle directory.
   Ho preferito farne una funzione piuttosto che ripetere piu' volte lo stesso codice.
*/

const int analyze_cases(const int var)
{
    switch(var)
    {
        case WALK_OK: break;
        case WALK_NAMETOOLONG: puts("Filename or directory name too long!"); break;
        case WALK_OPEN_ERROR:  puts("Can't open the source directory"); break;
        case WALK_REMOVE_FILE: puts("Error while removing the last file!"); break;
    }
    return 0;
}





/* Funzione d'utilita' che controlla i diritti d'accesso ad un file e lo elimina */

const int check_and_remove(const char *namesrc)
{
    if (!isDirectory(namesrc)) remove_error(remove(namesrc));
    else analyze_cases(remove_recur(namesrc));
    return 0;
}





/* Funzione d'utilita' che comunica l'invalidita' dell'argomento passato */

const inline void invalid_argument(const char *namesrc)
{
    printf("invalid argument : %s",namesrc);
}



 
/* Funzione d'utilita' che restituisce la directory corrente salvandola in un buffer tramite getcwd */

const inline char *current_directory()
{
    char buffer[MAX_BUFFER_SIZE];
    return getcwd(buffer,MAX_BUFFER_SIZE);
}



/* Controlla l'effettiva allocazione di memoria dinamica */

const inline void check_alloc(const void *ptr)
{
    if (ptr==NULL)
    {
        puts(BAD_ALLOC);
        exit(EXIT_FAILURE);
    }
}




/* Funzione d'utilita' che stampa le informazioni su un file */

const void info_file(const char *name_file)
{
    struct stat info;
    if (stat(name_file,&info)!=-1)
    {
        printf("\n\n==> %s <==",name_file);
        if (isDirectory(name_file)) puts("\n\nType: DIRECTORY");
        else puts("Type: FILE");
        printf("Number of inode:        %llu\n",info.st_ino);
        printf("Protection rights:      %hu\n",info.st_mode);
        printf("Hard links number:      %d\n",info.st_nlink);
        printf("ID User:                %u\n",info.st_uid);
        printf("Bytes length:           %lld\n",info.st_size);
        printf("Filesystem block size:  %d\n",info.st_blksize);
        printf("Blocks occupied:        %lld\n",info.st_blocks);
        printf("Last access:            %s",ctime(&info.st_atime));
        printf("Last file change:       %s",ctime(&info.st_mtime));
        printf("Last data change:       %s",ctime(&info.st_ctime));
    }
    else printf("Can't obtain information about: %s",name_file);
}